# -*- coding:utf-8 -*-
import time
import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import DataLoader, Dataset
from tqdm import tqdm
import os
from torch.utils import tensorboard
from sklearn.metrics import  accuracy_score,roc_auc_score
import copy
import json
import random
from torch.cuda.amp import autocast,GradScaler
from matplotlib import pyplot as plt
import seaborn as sns
import pandas as pd

class Metrics(object):
    def __init__(self):
        self.acc = self.accuracy
        self.roc_auc = roc_auc_score

    def accuracy(self,y_preds,y_true):
        y_preds = np.argmax(y_preds,axis=-1)
        y_true = y_true
        return accuracy_score(y_pred=y_preds,y_true=y_true)


class Conifg(object):
    def __init__(self):
        #general_dict | train_dict | checkpoint_dict | model_dict
        self.create_config()

    def general_config(self):
        general_dict = {}
        general_dict.update(
            {
            'seed':0,
            'use_multi_gpu' : False,
            'use_gpu': True,
            'log_dir' : 'monitor',
            'experiment_name' : "auto",
            'output_train_log':True,
            'output_visual_graphs':True
            }
        )
        return general_dict

    def train_config(self):
        train_dict = {}
        train_dict.update({
            'accumulate_num': -1,
            'resume_path':None,
            'use_fp16': False,
        }
        )
        return train_dict

    def checkpoint_config(self):
        checkpoint_dict = {}
        checkpoint_dict.update( {
            'use_checkpoint':False,
            'monitor_by_train_loss':False,
            'monitor_by_val_loss':False,
            'monitor_by_train_metric':False,
            'monitor_by_val_metric':True,
            'use_early_stopping':False,
            'monitor_max':True,
            'save_top_k':-1,
            'save_last':True
        })
        return checkpoint_dict

    def model_config(self):
        model_dict = {
            'use_gradient_checkpoint': False,
        }
        return model_dict

    def create_config(self):
        self.general_dict = self.general_config()
        self.train_dict = self.train_config()
        self.model_dict = self.model_config()
        self.checkpoint_dict = self.checkpoint_config()

    def to_json(self,path):
        dicts = ['一般配置',self.general_dict,'训练参数配置',self.train_dict,'模型参数配置',self.model_dict,'模型检查点配置',self.checkpoint_dict]
        with open(os.path.join(path,'total_args.json'), 'w', encoding='utf-8') as json_file:
            json_file.write(json.dumps(dicts,ensure_ascii=False,indent=2))

config = Conifg()

class Trainer(object):
    def __init__(self,
                 config,
                 **kwargs
                 ):
        self.metric = None
        self.config = config


    def log(self,tag, scalar_value, global_step=None):
        self.writer.add_scalar(tag,scalar_value,global_step)

    def fit(self, model,train_loader: DataLoader,criterion,val_loader: DataLoader=None,metric=None,
            optimizer=None,schedule=None,
            num_epochs=None,num_steps=None,
            ):
        assert num_epochs is not  None or num_steps is not None
        assert optimizer is not None and train_loader is not None

        #init
        self.fit_init()

        self.num_epochs = num_epochs
        self.actual_accumulate_num = 0 if self.config.train_dict['accumulate_num'] > 0 else -1
        self.actual_step  = 0
        self.last_train_loss = np.inf
        self.last_val_loss = np.inf
        if self.config.general_dict['output_visual_graphs']:
            self.train_loss_history = []
            self.val_loss_history = []
            self.train_metric_history = []
            self.val_metric_history = []

        if self.config.checkpoint_dict['monitor_max']:
            self.last_train_metric = 0.
            self.last_val_metric = 0.
        else:
            self.last_train_metric = np.inf
            self.last_val_metric = np.inf

        self.metric = metric
        self.model = model
        self.optimizer = optimizer
        self.criterion = criterion
        if self.config.train_dict['use_fp16']:
            print('Use FP16!!!')
            self.scaler = GradScaler()


        for i in range(1,num_epochs + 1):
            one_time = time.time()
            print(f'\n***** Running training epoch {i} *****')
            train_avg_loss,train_metric_value = self.train_per_epoch(dataloader=train_loader)
            self.log('Train_AvgLoss_Epoch',train_avg_loss,i)
            if train_metric_value is not None:
                self.log(f'Train_{self.metric.__name__}', train_metric_value, i)
            print(f'\n***** Running valid epoch {i} *****')
            val_avg_loss,val_metric_value = self.val_per_epoch(dataloader=val_loader)
            self.log('Val_AvgLoss_Epoch',val_avg_loss,i)
            if val_metric_value is not None:
                self.log(f'val_{self.metric.__name__}', val_metric_value, i)

            if self.config.general_dict['output_visual_graphs']:
                self.train_loss_history.append(train_avg_loss)
                self.val_loss_history.append(val_avg_loss)
                self.train_metric_history.append(train_metric_value)
                self.val_metric_history.append(val_metric_value)

            two_time = time.time()
            self.epoch_information = f'Epoch_{i}'
            self.train_loss_infor = f'train_loss_{round(train_avg_loss,4)}'
            self.val_loss_infor = f'val_loss_{round(val_avg_loss,4)}'
            self.train_metric_infor = f'train_{self.metric.__name__}_{round(train_metric_value,4)}'
            self.val_metric_infor = f'val_{self.metric.__name__}_{round(val_metric_value,4)}'

            self.monitor_information = f'Epoch {i} : train_loss {round(train_avg_loss,4)}  | val_loss {round(val_avg_loss,4)} || train_{self.metric.__name__} {round(train_metric_value,4)} | val_{self.metric.__name__} {round(val_metric_value,4)} | Time {round(two_time-one_time,5)}s | lr = {self.optimizer.state_dict()["param_groups"][0]["lr"]} '
            #print
            print()
            print(self.monitor_information)
            if self.config.general_dict['output_train_log']:
                self.train_log.write(self.monitor_information+'\n')
            #checkpoint
            if self.config.checkpoint_dict['use_checkpoint']:
                self.save_checkpoint(train_avg_loss,val_avg_loss,train_metric_value,val_metric_value)

        #output_visual_graphs
        if self.config.general_dict['output_visual_graphs']:
            index = list(range(1,self.num_epochs+1))
            data = self.train_loss_history
            wide_df = pd.DataFrame(data, index, [str('best_'+self.train_loss_infor)])
            fig = sns.lineplot(data=wide_df)
            plt.xlabel("Epoch");plt.ylabel("Train_loss");plt.title("Visual Graph")
            fig.get_figure()
            plt.savefig(os.path.join(self.experiment_path,'train_loss.png'), dpi=400)
            plt.clf()

            data = self.val_loss_history
            wide_df = pd.DataFrame(data, index, [str('best_'+self.val_loss_infor)])
            fig = sns.lineplot(data=wide_df)
            plt.xlabel("Epoch");plt.ylabel("Val_loss");plt.title("Visual Graph")
            fig.get_figure()
            plt.savefig(os.path.join(self.experiment_path,'val_loss.png'), dpi=400)
            plt.clf()

            data = self.train_metric_history
            wide_df = pd.DataFrame(data, index, [str('best_'+self.train_metric_infor)])
            fig = sns.lineplot(data=wide_df)
            plt.xlabel("Epoch");plt.ylabel("train_"+self.metric.__name__);plt.title("Visual Graph")
            fig.get_figure()
            plt.savefig(os.path.join(self.experiment_path,str(f"train_{self.metric.__name__}.png")), dpi=400)
            plt.clf()

            data = self.val_metric_history
            wide_df = pd.DataFrame(data, index, [str('best_'+self.val_metric_infor)])
            fig = sns.lineplot(data=wide_df)
            plt.xlabel("Epoch");plt.ylabel("val_"+self.metric.__name__);plt.title("Visual Graph")
            fig.get_figure()
            plt.savefig(os.path.join(self.experiment_path,str(f"val_{self.metric.__name__}.png")), dpi=400)
            plt.clf()

        if self.config.checkpoint_dict['use_checkpoint'] and self.config.checkpoint_dict['save_last']:
            self.save_last()
        self.train_log.close()

    def train_per_step(self,batch):

        if self.config.general_dict['use_gpu']:
            for i,b in enumerate(batch):
                batch[i] = b.cuda()
        else:
            for i, b in enumerate(batch):
                batch[i] = b

        features,label = batch[:-1],batch[-1]

        if self.config.train_dict['use_fp16']:
            with autocast():
                y_pred = self.model(*features)
                loss = self.criterion(y_pred,label)
            if self.config.train_dict['accumulate_num'] > 0:
                loss = loss / self.config.train_dict['accumulate_num']
            self.scaler.scale(loss).backward()
        else:
            y_pred = self.model(*features)
            loss = self.criterion(y_pred, label)
            if self.config.train_dict['accumulate_num'] > 0:
                loss = loss / self.config.train_dict['accumulate_num']
            loss.backward()
        ##累积梯度

        self.accumulate_gradient()

        return loss.item(), y_pred, label

    def val_per_step(self,batch):

        if self.config.general_dict['use_gpu']:
            for i,b in enumerate(batch):
                batch[i] = b.cuda()
        else:
            for i, b in enumerate(batch):
                batch[i] = b

        features,label = batch[:-1],batch[-1]


        y_pred = self.model(*features)


        loss = self.criterion(y_pred,label)

        return loss.item(), y_pred, label

    def train_per_epoch(self,dataloader: DataLoader,
                        use_tqdm=True):
        self.steps_per_epoch = 0
        train_losses  = 0.
        y_preds = []
        labels = []

        if use_tqdm:
            loader_bar = tqdm(dataloader)
        else:
            loader_bar = dataloader

        self.model.train()
        for i,batch in enumerate(loader_bar):
            train_loss, y_pred, label = self.train_per_step(batch)
            labels.append(label.detach().cpu().numpy())
            y_preds.append(y_pred.detach().cpu().numpy())


            loader_bar.set_postfix({'train_loss':train_loss})
            self.log('train_loss_step',train_loss,self.actual_step)
            train_losses += train_loss
            self.steps_per_epoch +=1
            self.actual_step +=1

        train_losses = train_losses / self.steps_per_epoch

        y_preds = np.concatenate(y_preds, axis=0)
        labels = np.concatenate(labels, axis=None)

        if self.metric is not None:
            metric_value = self.metric(y_preds,labels)
        else:
            metric_value = None
        return train_losses,metric_value

    def val_per_epoch(self,dataloader: DataLoader,use_tqdm=True):
        self.steps_per_epoch = 0
        val_losses = 0
        y_preds = []
        labels = []
        if dataloader is None:
            return None,None
        if use_tqdm:
            loader_bar = tqdm(dataloader)
        else:
            loader_bar = dataloader

        self.model.eval()
        for i, batch in enumerate(loader_bar):
            with torch.no_grad():
                val_loss, y_pred, label= self.val_per_step(batch)
                loader_bar.set_postfix({'val_loss':val_loss})

            labels.append(label.detach().cpu().numpy())
            y_preds.append(y_pred.detach().cpu().numpy())

            val_losses += val_loss
            self.steps_per_epoch += 1

        val_outputs = val_losses / self.steps_per_epoch



        y_preds = np.concatenate(y_preds,axis=0)
        labels = np.concatenate(labels,axis=None)

        if self.metric is not None:
            metric_value = self.metric(y_preds,labels)
        else:
            metric_value = None

        return val_outputs, metric_value

    def savemodel(self,model_name):
        model_statedict = copy.deepcopy(self.model.state_dict())
        save_path = os.path.join(self.experiment_path,'checkpoint')
        if not os.path.exists(save_path):
            os.mkdir(save_path)
        torch.save(model_statedict,os.path.join(save_path,f'{model_name}.pth'))
        model_path = str(os.path.join(save_path,f'{model_name}.pth'))
        print(f'Save Model at {model_path}')

    def save_checkpoint(self,train_avg_loss,val_avg_loss,train_metric_value,val_metric_value):
        if self.config.checkpoint_dict['monitor_by_train_loss']:
            if train_avg_loss < self.last_train_loss:
                print(f'train_loss由{round(self.last_train_loss, 4)}降到{train_avg_loss}')
                if self.config.general_dict['output_train_log']:
                    self.train_log.write(f'train_loss由{round(self.last_train_loss, 4)}降到{train_avg_loss}' + '\n')
                self.last_train_metric = train_avg_loss
                self.savemodel(self.epoch_information + '_' + self.train_loss_infor)
        elif self.config.checkpoint_dict['monitor_by_val_loss']:
            if val_avg_loss < self.last_val_loss:
                print(f'val_loss由{round(self.last_val_loss, 4)}降到{val_avg_loss}')
                if self.config.general_dict['output_train_log']:
                    self.train_log.write(f'val_loss由{round(self.last_val_loss, 4)}降到{val_avg_loss}' + '\n')
                self.last_val_metric = val_avg_loss
                self.savemodel(self.epoch_information + '_' + self.val_loss_infor)
        elif self.config.checkpoint_dict['monitor_by_train_metric']:
            if self.config.checkpoint_dict['monitor_max']:

                if train_metric_value > self.last_train_metric:
                    print(
                        f'train_{self.metric.__name__}由{round(self.last_train_metric, 4)}提高到{round(train_metric_value, 4)}')
                    if self.config.general_dict['output_train_log']:
                        self.train_log.write(f'train_{self.metric.__name__}由{round(self.last_train_metric, 4)}提高到{round(train_metric_value, 4)}' + '\n')
                    self.last_train_metric = train_metric_value
                    self.savemodel(self.epoch_information + '_' + self.train_metric_infor)
            else:
                if train_metric_value < self.last_train_metric:
                    print(
                        f'train_{self.metric.__name__}由{round(self.last_train_metric, 4)}降低到{round(train_metric_value, 4)}')
                    if self.config.general_dict['output_train_log']:
                        self.train_log.write(f'train_{self.metric.__name__}由{round(self.last_train_metric, 4)}降低到{round(train_metric_value, 4)}' + '\n')
                    self.last_train_metric = train_metric_value
                    self.savemodel(self.epoch_information + '_' + self.train_metric_infor)
        elif self.config.checkpoint_dict['monitor_by_val_metric']:
            if self.config.checkpoint_dict['monitor_max']:

                if val_metric_value > self.last_val_metric:
                    print(
                        f'val_{self.metric.__name__}由{round(self.last_val_metric, 4)}提高到{round(val_metric_value, 4)}')
                    if self.config.general_dict['output_train_log']:
                        self.train_log.write(f'val_{self.metric.__name__}由{round(self.last_val_metric, 4)}提高到{round(val_metric_value, 4)}' + '\n')
                    self.last_val_metric = val_metric_value
                    self.savemodel(self.epoch_information + '_' + self.val_metric_infor)
            else:
                if val_metric_value < self.last_val_metric:
                    print(
                        f'train_{self.metric.__name__}由{round(self.last_val_metric, 4)}降低到{round(val_metric_value, 4)}')

                    if self.config.general_dict['output_train_log']:
                        self.train_log.write(f'val_{self.metric.__name__}由{round(self.last_val_metric, 4)}降低到{round(val_metric_value, 4)}' + '\n')
                    self.last_val_metric = val_metric_value
                    self.savemodel(self.epoch_information + '_' + self.val_metric_infor)

    def accumulate_gradient(self):
        if self.config.train_dict['accumulate_num'] < 0 or self.actual_accumulate_num == self.config.train_dict['accumulate_num'] -1:
            if self.config.train_dict['use_fp16']:
                self.scaler.step(self.optimizer)
                self.scaler.update()
                self.optimizer.zero_grad()
            else:
                self.optimizer.step()
                self.optimizer.zero_grad()
            if self.config.train_dict['accumulate_num'] > 0:
                self.actual_accumulate_num = 0
        else:
            if self.config.train_dict['accumulate_num']> 0:
                self.actual_accumulate_num += 1

    def save_last(self):
        if self.config.checkpoint_dict['monitor_by_train_loss']:
            self.savemodel('last_model_'+ self.epoch_information + '_' + self.train_loss_infor)
        elif self.config.checkpoint_dict['monitor_by_val_loss']:
            self.savemodel('last_model_'+ self.epoch_information + '_' + self.val_loss_infor)
        elif self.config.checkpoint_dict['monitor_by_train_metric']:
            self.savemodel('last_model_'+ self.epoch_information + '_' + self.train_metric_infor)
        elif self.config.checkpoint_dict['monitor_by_val_metric']:
            self.savemodel('last_model_'+ self.epoch_information + '_' + self.val_metric_infor)

    def fit_init(self):


        if config.general_dict['log_dir'] is not None:
            if not os.path.exists(config.general_dict['log_dir']):
                os.mkdir(config.general_dict['log_dir'])
        self.experiment_id = 0
        if config.general_dict['experiment_name'] == 'auto':
            while os.path.exists(os.path.join(config.general_dict['log_dir'],"exp"+str(self.experiment_id))):
                self.experiment_id +=1

            self.experiment_path = os.path.join(config.general_dict['log_dir'], "exp"+str(self.experiment_id))
            os.mkdir(self.experiment_path)
        else:
            self.experiment_path = os.path.join(config.general_dict['log_dir'], config.general_dict['experiment_name'])
            if not os.path.exists(self.experiment_path):
                os.mkdir(self.experiment_path)

        self.writer  = tensorboard.SummaryWriter(log_dir=self.experiment_path,comment='trainer')
        self.config.to_json(self.experiment_path)
        if self.config.general_dict['output_train_log']:
            self.train_log = open(os.path.join(self.experiment_path,'train_log.txt'),'w',encoding='utf-8')


def seed_everything(seed,deterministic=False):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    if deterministic:
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False

def distributed_learning_init():
    pass


